@@ -267,16 +267,15 @@
 .http_200 = function(socket, content_type="text/html; charset=UTF-8",
                     content="<html><body><h1>R Websocket Server</h1></body></html>")
 {
-  n = ifelse(is.character(content),nchar(content), length(content))
+  if (is.character(content)) content = charToRaw(content)
+  n = length(content)
+
   h="HTTP/1.1 200 OK\r\nServer: R/Websocket\r\n"
   h=paste(h,"Content-Type: ",content_type, "\r\n",sep="")
   h=paste(h,"Date: ",date(),"\r\n",sep="")
   h=paste(h,"Content-Length: ",n,"\r\n\r\n",sep="")
   .SOCK_SEND(socket,charToRaw(h))
-  if(is.character(content))
-    .SOCK_SEND(socket,charToRaw(content))
-  else
-    .SOCK_SEND(socket,content)
+  .SOCK_SEND(socket,content)
   .SOCK_CLOSE(socket)
   TRUE
 }
